import os
import sys
import urllib
import urllib2
import base64
import requests

from servicemanager.smfile import remove_if_exists
from actions.colours import BColors
from smnexus import SmNexus
from xml.dom.minidom import parse


b = BColors()


class SmBintray():

    def __init__(self, context, service_name):
        self.context = context
        self.service_name = service_name
        self.service_type = context.service_type(service_name)
   
    def _find_latest_in_dom(self, dom):
        try:
            data = dom.getElementsByTagName("versioning")[0]
        except:
            self.context.log("Unable to get latest version from bintray")
            return None

        latestVersion =  data.getElementsByTagName("latest")[0].firstChild.nodeValue
        return latestVersion

    def _get_version_info_from_bintray(self, artifact, repositoryId, groupId):
        url = self.context.config_value("bintray")["protocol"] + "://" + self.context.config_value("bintray")["host"] + "/" + repositoryId + "/" + groupId + artifact + "/maven-metadata.xml"
        request = urllib2.Request(url)
        response = urllib2.urlopen(request)
        dom = parse(response)
        response.close()
        return self._find_latest_in_dom(dom)

    def find_latest_version(self, run_from, artifact, groupId):    
        version_env_var = None
        if "versionEnv" in self.context.service_data(self.service_name):
            version_env_var = self.context.service_data(self.service_name)["versionEnv"]

        try:
            version = os.environ[version_env_var]
        except Exception:
            repo_mappings = self.context.config_value("bintray")["repoMappings"]
            version = self._get_version_info_from_bintray(artifact, repo_mappings[run_from], groupId)
        return version

    def _download_from_bintray(self, bintray_path, local_filename, repositoryId, show_progress):
        url = self.context.config_value("bintray")["protocol"] + "://" + self.context.config_value("bintray")["host"] + "/" + repositoryId + "/" + bintray_path
        if show_progress:
            urllib.urlretrieve(url, local_filename, SmNexus._report_hook)
            print("\n")
        else:
            urllib.urlretrieve(url, local_filename)

    def download_jar_if_necessary(self, run_from, version):
        artifact = self.context.service_data(self.service_name)["binary"]["artifact"]
        groupId = self.context.service_data(self.service_name)["binary"]["groupId"]
        repo_mappings = self.context.config_value("bintray")["repoMappings"]
        repositoryId = repo_mappings[run_from]

        if not version:
            version = self.find_latest_version(run_from, artifact, groupId)

        if version:   

            localFilename = artifact + ".tgz"
            bintrayFilename = artifact + "-" + str(version) + ".tgz"
            bintrayFilePath = groupId + artifact + "/" + str(version) + "/" + bintrayFilename
            bintrayMD5FilePath = bintrayFilePath + ".md5"
            microservice_target_path = self.context.get_microservice_target_path(self.service_name)
            downloaded_artifact_path = microservice_target_path + localFilename
            downloaded_md5_path = microservice_target_path + localFilename + ".md5"

             #first download the md5 file in order to determine if new artifact download is required
            self._download_from_bintray(bintrayMD5FilePath, downloaded_md5_path, repositoryId, False)

            bintray_md5 = open(downloaded_md5_path, 'r').read()
            local_md5 = SmNexus._md5_if_exists(downloaded_artifact_path)

            if local_md5 != bintray_md5:
                remove_if_exists(downloaded_artifact_path)
                self.context.log("Downloading Bintray binary for '" + self.service_name + "': " + bintrayFilename)
                self._download_from_bintray(bintrayFilePath, downloaded_artifact_path, repositoryId, self.context.show_progress)
            os.remove(downloaded_md5_path)
        else:
            print b.warning + "WARNING: Due to lack of version data from Bintray you may not have an up to date version..." + b.endc    


#!/usr/bin/env python
import os

from actions.colours import BColors
from servicemanager import subprocess


b = BColors()


def pull_rebase_repo(context, name, project_info):
    if "sources" in project_info and "repo" in project_info["sources"]:
        print "pulling '" + name + "' from repo '" + project_info["sources"]["repo"] + "'" 
        path = context.application.workspace + project_info["location"]
        if not os.path.exists(path):
            print b.fail + "Nothing was pulled, it appears you have not cloned this repo yet: '" + name + "'" + b.endc + "\n"
        else:
            os.chdir(path)
            command = 'git pull --rebase'
            print "running: '" + command + "' from: '" + os.getcwd() + "'" 
            ps_command = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
            stdout, stderr = ps_command.communicate()
            if ps_command.returncode is not 0:
                print b.fail + "Nothing was pulled, see output for: '" + name + "'" + b.endc
            else:
                print b.okgreen + "Pulled: '" + name + "'" + b.endc
            print stdout


def clone_repo_if_required_raw(service_name, repo, path, context):
    path_exists = True
    if not os.path.exists(path):
        path_exists = False
        context.log("Source code for %s is missing, cloning..." % service_name)
        os.chdir(context.application.workspace)
        command = 'git clone %s' % repo
        ps_command = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        ps_command.communicate()
        if ps_command.returncode is not 0:
            print b.fail + "ERROR: Unable to clone repo for '" + service_name + "'" + b.endc
        if os.path.exists(path):
            path_exists = True
            print service_name + " - cloned"
        else:
            # TODO Should this just throw an exception?
            context.log("Could not go to dir: " + path + " do you have the project: " + service_name + " checked out?")
    elif os.path.exists(path + "/.git"):
        #TODO: make non os dependent
        print service_name + " - source exists"
    else:
        print b.warning + "WARNING: Nothing was cloned for '" + service_name + "' folder '" + path + "' already exists and it does not contain a repo" + b.endc

    return path_exists


def clone_repo_if_requred(service):

    context = service.context
    service_data = service.service_data
    sources_data = service.sources
    path = context.application.workspace + service_data["location"]

    return clone_repo_if_required_raw(service.service_name, sources_data["repo"], path, context)



class BColors:

    def __init__(self):
        self.bold = "\033[1m"
        self.okblue = '\033[94m'
        self.okgreen = '\033[92m'
        self.warning = '\033[93m'
        self.fail = '\033[91m'
        self.endc = '\033[0m'


import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../servicemanager'))

# dont do this in production code, this is bad practice it would seem, only for tests
from servicemanager.actions import actions
from servicemanager.server import smserverlogic
from servicemanager.smcontext import SmApplication, SmContext, ServiceManagerException
from servicemanager.smprocess import SmProcess
from servicemanager.service.smplayservice import SmPlayService
from servicemanager.server.smserverlogic import BadRequestException

import pytest
import time
import shutil
import unittest

from servicemanager.serviceresolver import ServiceResolver

from BaseHTTPServer import HTTPServer
from SocketServer import ThreadingMixIn
from servicemanager.smcredentials import EnvNexusCredentials, CredentialsResolver, SbtNexusCredentials

class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""


def set_up_and_clean_workspace():
    workspace_dir = os.path.join(os.path.dirname(__file__), "workspace")
    if os.path.exists(workspace_dir):
        shutil.rmtree(workspace_dir)
    os.mkdir(workspace_dir)
    os.environ["WORKSPACE"] = workspace_dir
    os.chdir(workspace_dir)


class TestFileServer(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_file_server(self):
        name = "FAKE_NEXUS"
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)

        context.kill(name)
        self.assertEqual(context.get_service(name).status(), [])
        time.sleep(2)

        response1 = actions.start_one(context, name, True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service(name).status())
        response2 = actions.start_one(context, name, True, False, None, port=None)
        self.assertFalse(response2)
        context.kill(name)
        self.assertEqual(context.get_service(name).status(), [])


class TestNexus(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_nexus(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")

        # start fake nexus
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        context = SmContext(SmApplication(config_dir_override), None, False, False)
        servicetostart = "PLAY_NEXUS_END_TO_END_TEST"
        actions.start_one(context, servicetostart, True, False, None, port=None)
        self.assertIsNotNone(context.get_service(servicetostart).status())
        context.kill(servicetostart)

        context.kill("FAKE_NEXUS")

        self.assertEqual(context.get_service(servicetostart).status(), [])
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])


class TestBintray(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_bintray(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")

        # start fake bintray
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        response1 = actions.start_one(context, "FAKE_BINTRAY", True, False, None, port=None)
        self.assertIsNotNone(context.get_service("FAKE_BINTRAY").status())
        time.sleep(5)

        context = SmContext(SmApplication(config_dir_override), None, False, False)
        servicetostart = "PLAY_BINTRAY_END_TO_END_TEST"
        actions.start_one(context, servicetostart, True, False, None, port=None)
        self.assertIsNotNone(context.get_service(servicetostart).status())
        context.kill(servicetostart)

        context.kill("FAKE_BINTRAY")

        self.assertEqual(context.get_service(servicetostart).status(), [])
        self.assertEqual(context.get_service("FAKE_BINTRAY").status(), [])


class TestActions(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_start_and_stop_one(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        actions.start_one(context, "TEST_ONE", True, False, None, port=None)
        self.assertEquals(len(context.get_service("TEST_ONE").status()), 1)
        context.kill("TEST_ONE")
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_start_and_stop_one_with_append_args(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        actions.start_one(context, "TEST_ONE", True, False, None, None, ["; echo 'Fin du sleep!!'"])
        self.assertEquals(len(context.get_service("TEST_ONE").status()), 2) # it is two in this case because the append creates a forked process
        context.kill("TEST_ONE")
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_dropwizard_from_source(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)

        servicetostart = "DROPWIZARD_NEXUS_END_TO_END_TEST"
        actions.start_and_wait(service_resolver, context, [servicetostart], False, False, None, port=None, seconds_to_wait=90, append_args=None)
        self.assertIsNotNone(context.get_service(servicetostart).status())
        context.kill(servicetostart)
        self.assertEqual(context.get_service(servicetostart).status(), [])

    def test_dropwizard_from_jar(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)

        # start fake nexus
        actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        servicetostart = "DROPWIZARD_NEXUS_END_TO_END_TEST"
        actions.start_and_wait(service_resolver, context, [servicetostart], True, False, None, port=None, seconds_to_wait=90, append_args=None)
        self.assertIsNotNone(context.get_service(servicetostart).status())
        context.kill(servicetostart)
        context.kill("FAKE_NEXUS")
        time.sleep(5)
        self.assertEqual(context.get_service(servicetostart).status(), [])
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])

    def test_play_from_source(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)

        servicetostart = "PLAY_NEXUS_END_TO_END_TEST"
        port = None
        secondsToWait = 90
        append_args = None
        actions.start_and_wait(service_resolver, context, [servicetostart], False, False, None, port, secondsToWait, append_args)
        self.assertIsNotNone(context.get_service(servicetostart).status())
        context.kill(servicetostart)
        self.assertEqual(context.get_service(servicetostart).status(), [])

    def test_successful_play_from_jar_without_waiting(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)

        context.kill_everything()
        time.sleep(5)

        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        fatJar = True
        release = False
        proxy = None
        port = None
        seconds_to_wait = None
        append_args = None

        try:
            servicetostart = ["PLAY_NEXUS_END_TO_END_TEST"]
            actions.start_and_wait(service_resolver, context, servicetostart, fatJar, release, proxy, port, seconds_to_wait, append_args)
        finally:
            context.kill_everything()

    def test_successful_play_from_jar_without_waiting_with_append_args(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)

        context.kill_everything()
        time.sleep(5)

        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, None, None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        servicetostart = ["PLAY_NEXUS_END_TO_END_TEST"]
        appendArgs = {"PLAY_NEXUS_END_TO_END_TEST": ["-DFoo=Bar"]}
        fatJar = True
        release = False
        proxy = None
        port = None
        seconds_to_wait = None

        try:
            actions.start_and_wait(service_resolver, context, servicetostart, fatJar, release, proxy, port, seconds_to_wait, appendArgs)
            time.sleep(5)
            service = SmPlayService(context, "PLAY_NEXUS_END_TO_END_TEST")
            processes = SmProcess.processes_matching(service.pattern)
            self.assertEqual(len(processes), 1)
            self.assertTrue("-DFoo=Bar" in processes[0].args)
        finally:
            context.kill_everything()

    def test_failing_play_from_jar(self):

        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)

        context.kill_everything()
        time.sleep(5)

        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        try:
            servicetostart = ["BROKEN_PLAY_PROJECT"]
            actions.start_and_wait(service_resolver, context, servicetostart, fatjar=True, release=False, proxy=None, port=None, seconds_to_wait=2, append_args=None)
            self.fail("Did not expect the project to startup.")
        except ServiceManagerException as sme:
            self.assertEqual("Timed out starting service(s): BROKEN_PLAY_PROJECT", sme.message)
        finally:
            context.kill_everything()


    def test_start_and_stop_one_duplicate(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        response1 = actions.start_one(context, "TEST_ONE", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("TEST_ONE").status())
        response2 = actions.start_one(context, "TEST_ONE", True, False, None, port=None)
        self.assertFalse(response2)
        context.kill("TEST_ONE")
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_assets_server(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()

        context.kill("FAKE_NEXUS")
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])
        time.sleep(2)

        # start fake nexus
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])
        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        actions.start_one(context, "PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND", True, False, None, port=None)
        self.assertIsNotNone(context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND").status())
        context.kill("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND")
        context.kill("FAKE_NEXUS")
        time.sleep(15)

        self.assertEqual(context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND").status(), [])
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])

    def test_wait_on_assets_server(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        sm_application = SmApplication(config_dir_override)
        context = SmContext(sm_application, None, False, False)
        service_resolver = ServiceResolver(sm_application)
        context.kill_everything()

        context.kill("FAKE_NEXUS")
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])
        time.sleep(2)

        # start fake nexus
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])
        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        port = None
        seconds_to_wait = 5
        append_args = None
        actions.start_and_wait(service_resolver, context, ["PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND"], True, False, None, port, seconds_to_wait, append_args)
        self.assertIsNotNone(context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND").status())
        context.kill("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND")
        context.kill("FAKE_NEXUS")
        time.sleep(15)

        self.assertEqual(context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND").status(), [])
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])

    def test_python_server_offline(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, True, False)
        port = None
        append_args = None
        actions.start_one(context, "PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND", True, False, None, port, append_args)
        self.assertIsNotNone(context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND").status())
        context.kill("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND")
        time.sleep(5)
        self.assertEqual(context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND").status(), [])

class TestStartCommands(unittest.TestCase):

    def setUp(self):
        set_up_and_clean_workspace()

    def test_play_binary_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("PLAY_NEXUS_END_TO_END_TEST", True, False, None, port=None)
        #starter = SmPlayServiceStarter(context, "PLAY_NEXUS_END_TO_END_TEST", True, False, None, None, None, None)
        expected = [ './basicplayapp/bin/basicplayapp',
                                '-DProd.microservice.whitelist.useWhitelist=false',
                                '-DProd.mongodb.uri=mongodb://localhost:27017/auth',
                                '-J-Xmx256m',
                                '-J-Xms256m',
                                '-J-XX:MaxPermSize=128m',
                                '-Dhttp.port=8500',
                                '-Dservice.manager.serviceName=PLAY_NEXUS_END_TO_END_TEST',
                                '-Dservice.manager.runFrom=True']
        self.assertEqual(starter.get_start_command("BINARY"), expected)

    def test_play_source_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("PLAY_NEXUS_END_TO_END_TEST", True, False, None, port=None)
        expected = [ 'play', 'start -Dhttp.port=8500 -Dservice.manager.serviceName=PLAY_NEXUS_END_TO_END_TEST -Dservice.manager.runFrom=True -DFoo=false']
        self.assertEqual(starter.get_start_command("SOURCE"), expected)

    def test_dropwizard_binary_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("DROPWIZARD_NEXUS_END_TO_END_TEST", "foo", proxy=None)
        expected = [
            'java',
            '-Dfile.encoding=UTF8',
            '-Xmx64M',
            '-XX:+CMSClassUnloadingEnabled',
            '-XX:MaxPermSize=64m',
            '-Ddw.http.port=8080',
            '-Dservice.manager.serviceName=DROPWIZARD_NEXUS_END_TO_END_TEST',
            '-Dservice.manager.serviceName=DROPWIZARD_NEXUS_END_TO_END_TEST',
            '-Dservice.manager.runFrom=foo',
            '-jar',
            'dwtest-foo-shaded.jar',
            'server',
            'dev_config.yml']
        cmd = starter.get_start_command("BINARY")
        cmd[-1] = cmd[-1].split("/")[-1]
        cmd[0] = cmd[0].split("/")[-1]
        cmd[len(cmd) -3] = cmd[len(cmd) -3].split("/")[-1]
        self.assertEqual(cmd, expected)

    def test_dropwizard_source_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("DROPWIZARD_NEXUS_END_TO_END_TEST", "foo", proxy=None)
        expected = ['./startappfromcode.sh']
        self.assertEqual(starter.get_start_command("SOURCE"), expected)

    def test_python_binary_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND", "foo", proxy=None)
        expected = [ 'get_start_command() not implemented for this type of service - fork and make a pull request :)' ]
        cmd = starter.get_start_command("BINARY")
        self.assertEqual(cmd, expected)

    def test_python_source_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND", "foo", proxy=None)
        expected = ['get_start_command() not implemented for this type of service - fork and make a pull request :)']
        self.assertEqual(starter.get_start_command("SOURCE"), expected)

    def test_external_binary_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("FAKE_NEXUS", "foo", proxy=None)
        expected = [ 'python', 'fakenexus.py']
        cmd = starter.get_start_command("BINARY") #context will be ignored
        self.assertEqual(cmd, expected)

    def test_external_source_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        starter = context.get_service_starter("FAKE_NEXUS", "foo", proxy=None)
        expected = [ 'python', 'fakenexus.py']
        self.assertEqual(starter.get_start_command("SOURCE"), expected) #context will be ignored

class TestServerFunctionality(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_simple(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()
        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))
        request = dict()
        request["testId"] = "foo"
        request["services"] = [{"serviceName": "TEST_ONE", "runFrom": "SNAPSHOT"},
                               {"serviceName": "DROPWIZARD_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"},
                               {"serviceName": "PLAY_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"}]
        smserverlogic.SmStartRequest(server, request, True, False).process_request()
        self.assertIsNotNone(context.get_service("TEST_ONE").status())
        # stop does not currently work for extern
        # smserverlogic.SmStopRequest(SERVER, request).process_request()
        context.kill_everything()
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_play_with_append_args(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()

        # Start up fake nexus first
        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))
        request = dict()
        request["testId"] = "foo"
        request["services"] = [{"serviceName": "PLAY_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT", "appendArgs": ["-Dfoo=bar"]}]
        smserverlogic.SmStartRequest(server, request, True, False).process_request()
        time.sleep(5)
        self.assertEqual(len(context.get_service("PLAY_NEXUS_END_TO_END_TEST").status()), 1)
        service = SmPlayService(context, "PLAY_NEXUS_END_TO_END_TEST")
        processes = SmProcess.processes_matching(service.pattern)
        self.assertEqual(len(processes), 1)
        self.assertTrue("-Dfoo=bar" in processes[0].args)
        context.kill_everything()
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_play_with_invalid_append_args(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()
        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))
        request = dict()
        request["testId"] = "foo"
        request["services"] = [{"serviceName": "PLAY_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT", "appendArgs": "-Dfoo=bar"}]
        with pytest.raises(BadRequestException):
            smserverlogic.SmStartRequest(server, request, True, False).process_request()

    def test_external_with_append_args(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()
        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))
        request = dict()
        request["testId"] = "foo"
        request["services"] = [{"serviceName": "TEST_ONE", "runFrom": "SNAPSHOT", "appendArgs": [";echo foo"]}]
        smserverlogic.SmStartRequest(server, request, True, False).process_request()
        self.assertIsNotNone(context.get_service("TEST_ONE").status())
        pattern = context.application.services["TEST_ONE"]["pattern"]
        processes = SmProcess.processes_matching(pattern)
        # stop does not currently work for extern
        # smserverlogic.SmStopRequest(SERVER, request).process_request()
        self.assertEqual(len(processes), 2) #we expect two proecesses to be spawned because of the appended command
        self.assertTrue(";echo" in processes[0].args or ";echo" in processes[1].args)
        context.kill_everything()
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_external_with_invalid_append_args(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()
        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))
        request = dict()
        request["testId"] = "foo"
        request["services"] = [{"serviceName": "TEST_ONE", "runFrom": "SNAPSHOT", "appendArgs": ";echo foo"}]
        with pytest.raises(BadRequestException):
            smserverlogic.SmStartRequest(server, request, True, False).process_request()

    def test_offline(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()
        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))
        request = dict()
        request["testId"] = "foo"
        request["services"] = [{"serviceName": "TEST_ONE", "runFrom": "SNAPSHOT"},
                               {"serviceName": "DROPWIZARD_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"},
                               {"serviceName": "PLAY_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"}]
        smserverlogic.SmStartRequest(server, request, True, False).process_request()
        self.assertIsNotNone(context.get_service("TEST_ONE").status())
        # stop does not currently work for extern
        # smserverlogic.SmStopRequest(SERVER, request).process_request()
        context.kill_everything()
        self.assertEqual(context.get_service("TEST_ONE").status(), [])
        request["testId"] = "foo2"
        smserverlogic.SmStartRequest(server, request, True, True).process_request()
        self.assertIsNotNone(context.get_service("TEST_ONE").status())
        # stop does not currently work for extern
        #smserverlogic.SmStopRequest(SERVER, request).process_request()
        context.kill_everything()
        self.assertEqual(context.get_service("TEST_ONE").status(), [])

    def test_ensure_multiple_instances_of_a_service_can_be_started_from_server(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.kill_everything()
        server = smserverlogic.SmServer(SmApplication(config_dir_override, None))

        # start fake nexus
        self.assertEqual(context.get_service("FAKE_NEXUS").status(), [])
        response1 = actions.start_one(context, "FAKE_NEXUS", True, False, None, port=None)
        self.assertTrue(response1)
        self.assertIsNotNone(context.get_service("FAKE_NEXUS").status())
        time.sleep(5)

        first_request = dict()
        first_request["testId"] = "multiple-instance-unit-test-1"
        first_request["services"] = [{"serviceName": "TEST_ONE", "runFrom": "SNAPSHOT"},
                                     {"serviceName": "DROPWIZARD_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"},
                                     {"serviceName": "PLAY_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"}]
        request = smserverlogic.SmStartRequest(server, first_request, True, False)
        request.process_request()

        time.sleep(5)
        self.assertEqual(len(context.get_service("TEST_ONE").status()), 1)
        self.assertEqual(len(context.get_service("DROPWIZARD_NEXUS_END_TO_END_TEST").status()), 1)
        self.assertEqual(len(context.get_service("PLAY_NEXUS_END_TO_END_TEST").status()), 1)

        second_request = dict()
        second_request["testId"] = "multiple-instance-unit-test-2"
        second_request["services"] = [{"serviceName": "TEST_ONE", "runFrom": "SNAPSHOT"},
                                      {"serviceName": "DROPWIZARD_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"},
                                      {"serviceName": "PLAY_NEXUS_END_TO_END_TEST", "runFrom": "SNAPSHOT"}]
        smserverlogic.SmStartRequest(server, second_request, True, False).process_request()
        time.sleep(5)
        self.assertEqual(len(context.get_service("TEST_ONE").status()), 2)
        self.assertEqual(len(context.get_service("DROPWIZARD_NEXUS_END_TO_END_TEST").status()), 2)
        self.assertEqual(len(context.get_service("PLAY_NEXUS_END_TO_END_TEST").status()), 2)

        # stop does not currently work for extern
        # smserverlogic.SmStopRequest(SERVER, request).process_request()
        context.kill_everything()
        self.assertEqual(context.get_service("TEST_ONE").status(), [])
        context.kill("FAKE_NEXUS")


class TestConfiguration(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        application = SmApplication(config_dir_override, None)
        self.assertEqual(len(application.services), 11)
        self.assertEqual(application.services["TEST_TEMPLATE"]["type"], "external")
        self.assertEqual(application.services["TEST_TEMPLATE"]["pattern"], "some.namespace=TEST_TEMPLATE")
        self.assertEqual(application.services["TEST_TEMPLATE"]["includeInStartAndStopAll"], False)
        self.assertEqual(application.services["TEST_TEMPLATE"]["some_parameter"], "value we expect")
        self.assertEqual(application.services["TEST_TEMPLATE"]["anestedvalue"]["anotherunique"], "value")
        self.assertEqual(application.services["TEST_TEMPLATE"]["anestedvalue"]["new"], "and this value")
        self.assertEqual(application.services["TEST_TEMPLATE"]["anestedvalue"]["override"], "will have this value")
        self.assertEqual(len(application.services["TEST_TEMPLATE"]), 8)


class TestServiceResolver(unittest.TestCase):
    def setUp(self):
        set_up_and_clean_workspace()

    def test_config(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        application = SmApplication(config_dir_override, None)
        service_resolver = ServiceResolver(application)
        nexus_wildcard = service_resolver.resolve_services("PLAY_NEXU*TEST")
        self.assertTrue("PLAY_NEXUS_END_TO_END_TEST" in nexus_wildcard)
        self.assertEqual(1, len(nexus_wildcard))

        all_services = service_resolver.resolve_services("*")
        self.assertTrue("TEST_ONE" in all_services)
        self.assertTrue("TEST_TWO" in all_services)
        self.assertTrue("TEST_THREE" in all_services)
        self.assertTrue("TEST_TEMPLATE" in all_services)
        self.assertTrue("DROPWIZARD_NEXUS_END_TO_END_TEST" in all_services)
        self.assertTrue("PLAY_NEXUS_END_TO_END_TEST" in all_services)
        self.assertTrue("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND" in all_services)
        self.assertEqual(11, len(all_services))

        test_profile = service_resolver.resolve_services("TEST")
        self.assertTrue("TEST_ONE" in test_profile)
        self.assertTrue("TEST_TWO" in test_profile)
        self.assertTrue("TEST_THREE" in test_profile)
        self.assertTrue("TEST_TEMPLATE" in test_profile)
        self.assertEqual(4, len(test_profile))

        tests_without_one = service_resolver.resolve_services_from_array(["TEST", "-TEST_ONE"])
        self.assertFalse("TEST_ONE" in tests_without_one)
        self.assertTrue("TEST_TWO" in tests_without_one)
        self.assertTrue("TEST_THREE" in tests_without_one)
        self.assertTrue("TEST_TEMPLATE" in tests_without_one)
        self.assertEqual(3, len(tests_without_one))

        tests_without_one = service_resolver.resolve_services_from_array(["TEST_WILD_CARD_NEGATED_PROFILE"])
        self.assertTrue("TEST_ONE" in tests_without_one)
        self.assertFalse("TEST_TWO" in tests_without_one)
        self.assertTrue("TEST_THREE" in tests_without_one)
        self.assertTrue("TEST_TEMPLATE" in tests_without_one)
        self.assertEqual(3, len(tests_without_one))

        nothing = service_resolver.resolve_services_from_array(["*", "-*"])
        self.assertEqual(0, len(nothing))

        test_template = service_resolver.resolve_services("TEST_WILD_CARD_PROFILE")
        self.assertTrue("TEST_TEMPLATE" in test_template)
        self.assertEqual(1, len(test_template))


class TestConfig(unittest.TestCase):
    def test_runfrom_override(self):
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        python_server = context.get_service("PYTHON_SIMPLE_SERVER_ASSETS_FRONTEND")
        run_from = context.get_run_from_service_override_value_or_use_default(python_server, "SHOULD_BE_OVERWRITTEN")
        self.assertEqual(run_from, "RELEASE")


class TestCredentialsResolver(unittest.TestCase):

    defaultEnv = os.environ.copy()

    def test_env_nexus_creds(self):
        os.environ.clear()
        os.environ.update(self.defaultEnv)
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        creds = EnvNexusCredentials(context)
        resolver = CredentialsResolver(context)

        os.environ.setdefault("NEXUS_PASS", "lame")
        os.environ.setdefault("NEXUS_USER", "lame_user")

        self.assertEqual(creds.load_creds(), ("lame_user", "lame"))
        self.assertEqual(resolver.resolve_nexus_credentials(), {"user": "lame_user", "password": "lame"})


    def test_env_nexus_creds_from_config(self):
        os.environ.clear()
        os.environ.update(self.defaultEnv)
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.application.config["nexusPasswordEnvironmentVar"] = "A_FREAKIN_PASS_ENV_VAR"
        context.application.config["nexusUserEnvironmentVar"] = "A_FREAKIN_USER_ENV_VAR"
        creds = EnvNexusCredentials(context)
        resolver = CredentialsResolver(context)

        os.environ.setdefault("A_FREAKIN_PASS_ENV_VAR", "lame_from_conf")
        os.environ.setdefault("A_FREAKIN_USER_ENV_VAR", "lame_user_from_conf")

        self.assertEqual(creds.load_creds(), ("lame_user_from_conf", "lame_from_conf"))
        self.assertEqual(resolver.resolve_nexus_credentials(), {"user": "lame_user_from_conf", "password": "lame_from_conf"})

    def test_sbt_nexus_creds_from_config(self):
        os.environ.clear()
        os.environ.update(self.defaultEnv)
        config_dir_override = os.path.join(os.path.dirname(__file__), "conf")
        context = SmContext(SmApplication(config_dir_override), None, False, False)
        context.application.config["sbtCredentialsFile"] = os.path.dirname(__file__) + "/.sbt/.credentials"

        creds = SbtNexusCredentials(context)
        resolver = CredentialsResolver(context)

        self.assertEqual(creds.load_creds(), ("sbt", "sbt_creds"))
        self.assertEqual(resolver.resolve_nexus_credentials(), {"user": "sbt", "password": "sbt_creds"})


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import glob

import os
import re
import shutil
import zipfile
import tarfile
import stat
import copy
import types

from servicemanager.subprocess import Popen
from ..service.smservice import SmMicroServiceStarter
from smjvmservice import SmJvmService, SmJvmServiceStarter
from ..smfile import force_chdir, force_pushdir, remove_if_exists, remove_folder_if_exists, makedirs_if_not_exists
from ..smartifactrepofactory import SmArtifactRepoFactory
from ..actions.colours import BColors

from servicemanager import subprocess


b = BColors()


class SmPlayServiceStarter(SmJvmServiceStarter):

    PLAY_PROCESS_STARTUP_TIMEOUT_SECONDS = 120

    def _build_extra_params(self):
        extra_params = ["-Dhttp.port=%d" % self.port]
        extra_params += self.process_arguments()
        # Features are so specific - should this be in config?
        if self.context.features:
            for feature in self.context.features:
                extra_params += ["-Dfeature.%s=true" % feature]

        service_config_key = "Dev.microservice.services"

        if self.context.is_test:
            if self.service_data.get("hasMongo", False):
                extra_params += ["-DDev.microservice.mongodb.uri=mongodb://localhost:27017/%s-%s" % (self.context.database_name_prefix, self.service_name)]

        if self.service_mapping_ports and self.service_data.get("hasServiceMappings", False):
            for dependent_service_name in self.service_mapping_ports:
                service_config_key_with_prefix = service_config_key
                if self.service_name == "FEGOVUK":
                    service_config_key_with_prefix = "govuk-tax.Dev.services"
                extra_params += [
                    "-D%s.%s.host=localhost" % (service_config_key_with_prefix, dependent_service_name),
                    "-D%s.%s.port=%d" % (service_config_key_with_prefix, dependent_service_name, self.service_mapping_ports[dependent_service_name])
                ]

        if self.proxy:
            proxy_config = self.proxy.split(":")
            print "Starting service with proxy, '" + str(self.proxy) + "'"
            extra_params += [
                "-Dhttp.proxyHost=" + proxy_config[0],
                "-Dhttp.proxyPort=" + proxy_config[1]
            ]

        if self.append_args:
            if not isinstance(self.append_args, types.ListType):
                self.log("WARNING: I was passed a non list for append args of '" + str(self.append_args) + "' I dont know what to do with this")
            else:
                extra_params += self.append_args

        return extra_params

    def __init__(self, context, service_name, run_from, port, classifier, service_mapping_ports, version, proxy, append_args):
        SmMicroServiceStarter.__init__(self, context, service_name, "play", run_from, port, classifier, service_mapping_ports, version, proxy, append_args)

        if not self.port:
            self.port = self.service_data["defaultPort"]

    def supports_append_args(self):
        return True

    def get_start_command(self, run_from):
        if run_from == "SOURCE":
            source_cmd = copy.copy(self.service_data["sources"]["cmd"])
            source_cmd[-1] = source_cmd[-1] + " " +  " ".join(self.sbt_extra_params())
            return source_cmd
        else:
            return self.service_data["binary"]["cmd"] + self._build_extra_params()

    def start_from_binary(self):
        microservice_target_path = self.context.get_microservice_target_path(self.service_name)
        force_chdir(microservice_target_path)

        if not self.context.offline:
            binaryConfig = self.service_data["binary"]
            artifactRepo = SmArtifactRepoFactory.get_repository(self.context, self.service_name, binaryConfig)
            if not self.version:
                self.version = artifactRepo.find_latest_version(self.run_from, binaryConfig["artifact"], binaryConfig["groupId"])
            artifactRepo.download_jar_if_necessary(self.run_from, self.version)

        unzip_dir = self._unpack_play_application(SmArtifactRepoFactory.get_play_app_extension(binaryConfig))
        parent, _ = os.path.split(unzip_dir)
        force_pushdir(parent)

        if "frontend" in self.service_data and self.service_data["frontend"]:
           assets_versions = self._get_assets_version(unzip_dir)
           self.context.assets_versions_to_start(assets_versions)

        cmd_with_params = self.get_start_command("BINARY")
        if os.path.exists(cmd_with_params[0]):
            os.chmod(cmd_with_params[0], stat.S_IRWXU)
        else:
            print b.fail + "ERROR: unable to chmod on non existent file '" + parent + cmd_with_params[0] + "'" + b.endc

        makedirs_if_not_exists("logs")

        print(cmd_with_params)

        with open("logs/stdout.txt", "wb") as out, open("logs/stderr.txt", "wb") as err:
            popen_output = Popen(cmd_with_params, env=os.environ.copy(), stdout=out, stderr=err, close_fds=True)
            if popen_output.returncode == 1:
                print b.fail + "ERROR: could not start '" + self.service_name + "' " + b.endc
            return popen_output.pid

    def _unpack_play_application(self, extension):
        service_data = self.service_data
        microservice_zip_path = self.context.application.workspace + service_data["location"] + "/target/"
        force_pushdir(microservice_zip_path)
        zip_filename = service_data["binary"]["artifact"] + extension

        unzipped_dir = SmPlayService.unzipped_dir_path(self.context, service_data["location"])
        remove_folder_if_exists(unzipped_dir)

        os.makedirs(unzipped_dir)

        if extension == ".zip":
            self._unzip_play_application(zip_filename, unzipped_dir)
        elif extension == ".tgz":
            self._untar_play_application(zip_filename, unzipped_dir)
        else:
            print "ERROR: unsupported atrifact extension: " + extension         

        
        folder = [ name for name in os.listdir(unzipped_dir) if os.path.isdir(os.path.join(unzipped_dir, name)) ][0]
        target_dir = unzipped_dir + "/" + service_data["binary"]["destinationSubdir"]
        shutil.move(unzipped_dir + "/" + folder, target_dir)

        return target_dir

    def _unzip_play_application(self, zip_filename, unzipped_dir):
        zipfile.ZipFile(zip_filename, 'r').extractall(unzipped_dir)

    def _untar_play_application(self, tgz_filename, unzipped_dir):
        tfile = tarfile.open(tgz_filename, 'r:gz')
        tfile.extractall(unzipped_dir) 

    def sbt_extra_params(self):
        sbt_extra_params = self._build_extra_params()

        if "extra_params" in self.service_data["sources"]:
            sbt_extra_params += self.service_data["sources"]["extra_params"]

        return sbt_extra_params

    def start_from_sources(self):
        sbt_extra_params = self.sbt_extra_params()

        service_data = self.context.service_data(self.service_name)
        microservice_path = self.context.application.workspace + service_data["location"]
        curr_dir = force_pushdir(microservice_path)

        env_copy = os.environ.copy()
        env_copy["SBT_EXTRA_PARAMS"] = " ".join(sbt_extra_params) # TODO: not needed i think anymore...

        makedirs_if_not_exists("logs")

        with open("logs/stdout.txt", "wb") as out, open("logs/stderr.txt", "wb") as err:
            process = Popen(self.get_start_command("SOURCE"), env=env_copy, stdout=out, stderr=err, stdin=subprocess.PIPE)
            process.stdin.close()
            if process.returncode == 1:
                print b.fail + "ERROR: could not start '" + self.service_name + "' " + b.endc
            return process.pid # Note: This is the parent pid

    def _get_assets_version(self, unzip_dir):
        assets_versions = []
        for conf_file in glob.glob(unzip_dir + "/conf/*.conf"):
            with file(conf_file) as conf:
                conf = conf.read()
                conf_string = "".join(conf.split())
                pattern = re.compile(ur'Prod.*assets.*version="([0-9.]*)"')
                assets_versions = assets_versions + re.findall(pattern, conf_string)
        return assets_versions

class SmPlayService(SmJvmService):

    @staticmethod
    def unzipped_dir_path(context, location):
        return context.application.play_extraction_dir + location + "_" + context.instance_id

    def __init__(self, context, service_name):
        SmJvmService.__init__(self, context, service_name, "play")
        self.default_port = self.required_data("defaultPort")
        self.healthcheck = self.required_data("healthcheck")

    def post_stop(self):
        pass

    def clean_up(self):
        unzip_path = SmPlayService.unzipped_dir_path(self.context, self.service_data["location"])
        remove_folder_if_exists(unzip_path)

    def get_details_url(self):
        return "http://localhost:${port}/admin/details"

    def get_port_argument(self):
        return "http.port"

    def get_running_healthcheck_port(self, process):
        return process.extract_integer_argument('-D%s=(\d*)' % self.get_port_argument(), self.default_port)



#!/usr/bin/env python
import json
import time
import re
from abc import abstractmethod

import requests

from servicemanager.smprocess import SmProcess, kill_pid
from servicemanager.service.smservice import SmService, SmMicroServiceStarter, SmServiceStatus
from servicemanager.smrepo import clone_repo_if_requred


class SmJvmServiceStarter(SmMicroServiceStarter):

    test_id_param = "service.manager.testId"

    def __init__(self, context, service_name, expected_service_type, run_from, port, classifier, service_mapping_ports, version, proxy, append_args):
        SmMicroServiceStarter.__init__(self, context, service_name, expected_service_type, run_from, port, classifier, service_mapping_ports, version, proxy, append_args)

    def start(self, appendArgs=None):
        if self.run_from == "SOURCE":
            if not clone_repo_if_requred(self):
                # TODO - should this just return None? or throw an exception?  Should clone_repo_if_required throw an exception?
                return None

            return self.start_from_sources()
        else:
            return self.start_from_binary()

    def process_arguments(self):
        run_from = self.run_from
        if self.run_from=="RELEASE":
            run_from = (self.version or "UNKNOWN") + "-" + run_from
        jvm_args = ["-Dservice.manager.serviceName=%s" % self.service_name]
        jvm_args += ["-Dservice.manager.runFrom=%s" % run_from]
        if self.context.is_test:
            jvm_args += ["-Dservice.manager.testId=%s" % self.context.test_id]
            jvm_args += ["-Dservice.manager.startTime=%s" % time.time()]
        return jvm_args

    @abstractmethod
    def start_from_sources(self):
        pass

    @abstractmethod
    def start_from_binary(self):
        pass


class SmJvmService(SmService):

    def __init__(self, context, service_name, expected_service_type):
        SmService.__init__(self, context, service_name, expected_service_type)
        self.pattern = "service.manager.serviceName=%s$" % self.service_name
        self.default_port = self.required_data("defaultPort")
        self.healthcheck = self.required_data("healthcheck")

    @abstractmethod
    def post_stop(self):
        pass

    @abstractmethod
    def get_port_argument(self):
        pass

    @abstractmethod
    def get_running_healthcheck_port(self, process):
        pass

    @abstractmethod
    def get_details_url(self):
        pass

    def get_pattern(self):
        return self.pattern

    def status(self, all_processes=None):
        processes = SmProcess.processes_matching(self.pattern, all_processes)

        def _status_for_process(process):
            port = process.extract_integer_argument('-D%s=(\d*)' % self.get_port_argument(), self.default_port)
            test_id = process.extract_argument('-Dservice.manager.testId=([^ ]+)', "")
            run_from = process.extract_argument('-Dservice.manager.runFrom=([^ ]+)', "")
            features = process.extract_arguments('-Dfeature.([^ =]+)=true', "")
            healthcheck = SmServiceStatus.HEALTHCHECK_PASS if self.run_healthcheck(process) else SmServiceStatus.HEALTHCHECK_BOOT
            return SmServiceStatus.for_process(self.service_name, process, port, test_id, run_from, features, healthcheck)

        return map(_status_for_process, processes)

    def request_running_service_details_on_default_port(self):
        url = self.get_details_url().replace("${port}", str(self.default_port))

        try:
            response = requests.get(url)
            return json.loads(response.text)
        except requests.RequestException:
            return None

    def is_started_on_default_port(self):
        processes = SmProcess.processes_matching(self.pattern)
        default_port_argument = "-D%s=%d" % (self.get_port_argument, self.default_port)

        for process in processes:
            if process.has_argument(default_port_argument):
                return True

        return False

    def get_default_healthcheck_port(self):
        return self.default_port

    def stop(self):
        for process in SmProcess.processes_matching(self.pattern):
            kill_pid(self.context, process.ppid)
            kill_pid(self.context, process.pid)
            print "name: %s\tppid: %s\tpid: %s\tuptime: %s" % (self.service_name, process.ppid, process.pid, process.uptime)
        self.post_stop()

    def run_healthcheck(self, process):
        port = self.get_running_healthcheck_port(process)
        if not port:
            port = self.get_default_healthcheck_port()

        healthcheck_url = self.service_data["healthcheck"]["url"].replace("${port}", str(port))
        healthcheck_response_regex = self.service_data["healthcheck"]["response"] or ""

        try:
            ping_response = requests.get(healthcheck_url)
            response_text = ping_response.text
            return re.search(healthcheck_response_regex, response_text) and ping_response.status_code == 200
        except requests.RequestException:
            return False



import re


class ServiceResolver():

    def __init__(self, application):
        self.application = application

    def resolve_services(self, service_name):
        return self.resolve_services_from_array([service_name])

    def resolve_services_from_array(self, services):
        services_to_start = []
        services_to_not_start = []

        for service_name in services:
            if service_name.startswith('-'):
                services_to_not_start += self.resolve_services(service_name[1:])
            elif '*' in service_name:
                services_to_start += self._all_matching(service_name)
            elif self.application.has_service(service_name):
                services_to_start += [service_name]
            elif self.application.is_profile(service_name):
                services_to_start += self._get_all_in_profile(service_name)
            else:
                print "The requested service %s does not exist" % service_name
        for not_start in services_to_not_start:
            if not_start in services_to_start:
                services_to_start.remove(not_start)
        return services_to_start

    def _find_assets_service_name(self, service_name):
        assets_service_name = ''
        for service_name in self.application.services:
            if self.application.services[service_name]['type'] == 'assets':
                assets_service_name = service_name
        return assets_service_name

    def _get_all_in_profile(self, profile_name):
        services = []
        services_to_not_start = []

        assets_service_name = self._find_assets_service_name(profile_name)

        for service_name in self.application.services_for_profile(profile_name):
            if service_name.startswith('-'):
                services_to_not_start += self._all_matching(service_name[1:])
            elif '*' in service_name:
                services += self._all_matching(service_name)
            elif self.application.has_service(service_name):
                services.append(service_name)
            else:
                print "The requested service %s does not exist" % service_name
        for not_start in services_to_not_start:
            if not_start in services:
                services.remove(not_start)
        if assets_service_name in services:
            services.remove(assets_service_name)
            services.append(assets_service_name)
        return services

    def _all_matching(self, wildcard):
        services = []
        for service_name in self.application.services:
            if ServiceResolver._matches(service_name, wildcard):
                services.append(service_name)
        return services

    @staticmethod
    def _matches(service_name, wildcard):
        regex = re.compile(wildcard.replace('*', '.*'))
        return re.match(regex, service_name)


#!/usr/bin/env python

import json
import time
import requests
import sys
import traceback
import re
import types

from abc import abstractmethod
from bottle import request, response

from ..smcontext import SmContext, ServiceManagerException
from ..smport import PortProvider

RUN_ON_PORT = 8085
RUN_ON_HOST = 'localhost'
SERVICE_START_TIMEOUT_SECONDS = 90
MAX_TEST_ID_LENGTH = 40

deprecated_release_params = {"SNAPSHOT_JAR": "SNAPSHOT", "RELEASE_JAR": "RELEASE"}


class BadRequestException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class SmResponse:

    def __init__(self):
        pass  # Do nothing

    @staticmethod
    def bad_request(message):
        print "Bad Request: " + message
        response.status = 400
        return json.dumps({"statusCode": 400, "errorMessage": message})

    @staticmethod
    def error_500(message):
        response.status = 500
        return json.dumps({"statusCode": 500, "errorMessage": message})


class SmRequest:

    def __init__(self, server, json_body, offlineMode):
        self.server = server
        self.json_body = json_body

        try:
            test_id = self.json_body["testId"]
        except Exception:
            raise BadRequestException("Missing testId parameter")

        SmRequest._validate_test_id(test_id)

        request_specific_features = SmRequest._extract_and_validate_request_specific_features(self.json_body)

        self.test_id = test_id
        self.context = SmContext(server.application, self.test_id, request_specific_features=request_specific_features, offline=offlineMode)

    @abstractmethod
    def process_request(self):
        pass

    @staticmethod
    def _extract_and_validate_request_specific_features(json_body):

        if not "features" in json_body:
            return None

        request_specific_features = json_body["features"]

        if not request_specific_features:
            return None

        if not isinstance(request_specific_features, list):
            raise BadRequestException("'features' must be a list of strings")

        for feature in request_specific_features:
            if not isinstance(feature, basestring):
                raise BadRequestException("'features' must be a list of strings")

        return request_specific_features

    def _bad_request_exception(self, message):
        return BadRequestException("[%s] %s" % (self.test_id, message))

    def _log(self, message):
        self.context.log(message)

    @staticmethod
    def _validate_test_id(test_id):
        regex = re.compile("^[a-zA-Z0-9\-_]+$")

        if not regex.match(test_id):
            raise BadRequestException("Invalid parameter 'testId' with value '%s', valid characters are 'a-z', 'A-Z', '0-9', '-' and '_'" % test_id)

        if test_id.upper() == "LOCAL":
            raise BadRequestException("'%s' is not a valid value for testId" % test_id)

        if len(test_id) > MAX_TEST_ID_LENGTH:
            raise BadRequestException("Test id '%s' is too long (%d characters) (maximum is %d characters)" % (test_id, len(test_id), MAX_TEST_ID_LENGTH))

    def _get_or_throw_bad_request(self, obj, key, message):
        if key not in obj:
            raise self._bad_request_exception(message)

        value = obj[key]

        if not value:
            raise self._bad_request_exception(message)

        return value

    def _stop_services(self, drop_databases):

        self._log("Stopping services (drop databases = %s)" % drop_databases)

        errors = self.context.kill()

        if drop_databases:
            for service_name in self.server.service_names_for_test(self.test_id):
                if self.context.service_data(service_name).get("hasMongo", False):
                    self.context.drop_database_for_service(service_name)

            self.context.drop_database_for_test()

        self.server.test_stopped(self.test_id)

        return errors


class SmStartRequest(SmRequest):

    def __init__(self, server, json_request_body, do_not_run_from_source, offlineMode):
        self.do_not_run_from_source = do_not_run_from_source
        self.json_body = json_request_body
        SmRequest.__init__(self, server, self.json_body, offlineMode)

    def process_request(self):
        # START REQUEST PAYLOAD:
        #  {
        #   "testId": "blah",
        #   "features": ["feature1", "feature2", ...],
        #   "services": [
        #       {"serviceName" : "auth", "runFrom" : "SNAPSHOT"},
        #       {"serviceName" : "matching", "runFrom" : "RELEASE", "version" : "3.0.1"},
        #       {"serviceName" : "portal", "runFrom" : "SOURCE"},
        #       {"serviceName" : "nps",  "runFrom" : "SOURCE"},
        #       ...
        #   ]
        # }

        self._log("Processing service start request")

        if self.server.is_running(self.test_id):
            raise self._bad_request_exception("Test '%s' is already running" % self.test_id)

        self.server.starting_test(self.test_id)

        services_to_start = self._get_or_throw_bad_request(self.json_body, "services", "'services' missing from request")

        self._log("Service(s) to start: " + str(services_to_start))

        (orchestration_services, service_mapping_ports) = self._validate_start_request_and_assign_ports(services_to_start, self.do_not_run_from_source)

        try:
            self._start_services_for_test(orchestration_services, service_mapping_ports)

            sm_response = []

            for service_mapping_name in service_mapping_ports:
                sm_response += [{"serviceName": service_mapping_name, "port": service_mapping_ports[service_mapping_name]}]

            self._log("All services started! To kill the running processes for this test, POST {\"testId\":\"%s\"} to http://%s:%s/stop" % (self.test_id, RUN_ON_HOST, RUN_ON_PORT))

            return json.dumps(sm_response)

        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            return self._stop_services_and_return_500("Unexpected exception: " + e.message)

    # {"AUTH": {"port": 43124, "runFrom":"JAR", "serviceMapping" : "auth"}}
    def _start_services(self, orchestration_services, service_mapping_ports, proxy):

        for service_name in orchestration_services:

            port = orchestration_services[service_name]["port"]
            admin_port = orchestration_services[service_name]["adminPort"]
            run_from = orchestration_services[service_name]["runFrom"]
            classifier = orchestration_services[service_name]["classifier"]
            version = orchestration_services[service_name]["version"]
            append_args = orchestration_services[service_name]["appendArgs"] # Allows for dynamic config overriding

            # Allow for deprecated run_from values
            if run_from in deprecated_release_params:
                run_from = deprecated_release_params[run_from]

            self.context.start_service(service_name, run_from, proxy, classifier, service_mapping_ports, port, admin_port, version, append_args)

    def _await_service_startup(self, service_name, port, admin_port):
        seconds_remaining = SERVICE_START_TIMEOUT_SECONDS

        servicedata = self.context.service_data(service_name)
        if "healthcheck" in servicedata:
            healthcheck_url = servicedata["healthcheck"]["url"].replace("${port}", str(admin_port))
            healthcheck_response_regex = self.context.service_data(service_name)["healthcheck"]["response"]

            while seconds_remaining > 0:
                if (seconds_remaining < 10 or seconds_remaining % 5 == 0) and seconds_remaining != 1:
                    self._log("Waiting for %s to start on port %d, %d seconds before timeout" % (service_name, port, seconds_remaining))
                elif seconds_remaining == 1:
                    self._log("Waiting for %s to start on port %d, 1 second before timeout" % (service_name, port))

                try:
                    ping_response = requests.get(healthcheck_url)
                    response_text = ping_response.text
                    healthcheck = re.search(healthcheck_response_regex, response_text)
                except requests.RequestException:
                    healthcheck = False

                if healthcheck or (seconds_remaining == 0):
                    self._log("Service %s health check SUCCESSFUL" % service_name)
                    break
                else:
                    seconds_remaining -= 1
                    time.sleep(1)

                if seconds_remaining <= 0:
                    raise self.context.exception("Service %s - healthcheck did not pass within allocated time (%d seconds)" % (service_name, SERVICE_START_TIMEOUT_SECONDS))
        else:
            self._log("There is no health check for '%s'. This is not really advisable we can only assume it has started correctly" % service_name)

    def _start_services_for_test(self, orchestration_services, service_mapping_ports):

        self._start_services(orchestration_services, service_mapping_ports, None)

        for service_name in orchestration_services:

            self.server.starting_service_for_test(self.test_id, service_name)

            port = orchestration_services[service_name]["port"]
            admin_port = orchestration_services[service_name]["adminPort"]

            self._await_service_startup(service_name, port, admin_port)

    def _stop_services_and_return_500(self, message):
        self._log(message)
        errors = self._stop_services(drop_databases=True)

        if errors:
            self._log("Errors during shutdown: %s" % str(errors))

        return SmResponse.error_500(message)

    def _service_mapping_for(self, service_start_request):

        service_mapping_name = self._get_or_throw_bad_request(service_start_request, "serviceName", "Missing 'serviceName' parameter in instruction to start services")

        mapping = self._get_or_throw_bad_request(self.context.application.service_mappings, service_mapping_name, "Unknown service name '%s'" % service_mapping_name)

        need_classifier = isinstance(mapping, dict)
        have_classifier = "classifier" in service_start_request and service_start_request["classifier"]

        version = None
        if "version" in service_start_request and service_start_request["version"]:
            version = service_start_request["version"]

        append_args = service_start_request.get("appendArgs", [])

        if need_classifier:

            valid_classifiers = "[" + (",".join(str(x) for x in mapping.keys())) + "]"

            if not have_classifier:
                raise self._bad_request_exception("Service '%s' requires a classifier (one of: %s)" % (service_mapping_name, valid_classifiers))

            classifier = service_start_request["classifier"]

            if classifier not in mapping:
                raise self._bad_request_exception("Unknown classifier '%s' for service '%s' (expected one of: %s)" % (classifier, service_mapping_name, valid_classifiers))

            service_name = mapping[classifier]

        else:

            if have_classifier:
                raise self._bad_request_exception("Service '%s' does not take classifiers (found: '%s')" % (service_mapping_name, service_start_request["classifier"]))

            service_name = mapping
            classifier = None

        return service_mapping_name, service_name, classifier, version, append_args

    def _validate_start_request_and_assign_ports(self, services_to_start, dontrunfromsource):

        orchestration_services = {}
        service_mapping_ports = {}

        for service_start_request in services_to_start:

            service_mapping_name, service_name, classifier, version, append_args = self._service_mapping_for(service_start_request)

            if append_args and not isinstance(append_args, types.ListType):
                raise self._bad_request_exception("ERROR: I was passed a non list for append args of '" + str(append_args) + "' I dont know what to do with this")

            if service_mapping_name in service_mapping_ports:
                raise self._bad_request_exception("Duplicate entry for service '%s' in start request" % service_mapping_name)

            run_from = self._get_or_throw_bad_request(service_start_request, "runFrom", "Missing 'runFrom' parameter in instruction to start '%s'" % service_mapping_name)

            if run_from not in ["SOURCE", "SNAPSHOT", "RELEASE"] + deprecated_release_params.keys():
                raise self._bad_request_exception("runFrom parameter has invalid value '%s' (should be 'SOURCE', 'SNAPSHOT' or 'RELEASE')" % run_from)

            if dontrunfromsource:
                if run_from == "SOURCE":
                    raise self._bad_request_exception("runFrom parameter has value '%s', however --nosource was specified when smserver started" % run_from)

            if append_args and not self.context.get_service_starter(service_name, run_from, None).supports_append_args():
                raise BadRequestException("The service type for '" + service_name + "' does not support append args")

            if service_name in orchestration_services:
                existing_entry = orchestration_services[service_name]
                service_mapping_ports[service_mapping_name] = existing_entry["port"]

                if run_from != existing_entry["runFrom"]:
                    raise self._bad_request_exception("Conflicting runFrom values (%s and %s) for underlying service '%s'" % (run_from, existing_entry["runFrom"], service_name))

                if classifier and existing_entry["classifier"] and classifier != existing_entry["classifier"]:
                    raise self._bad_request_exception("Conflicting classifier values (%s and %s) for underlying service '%s'" % (classifier, existing_entry["classifier"], service_name))

            else:
                port = self.server.next_available_port()
                admin_port = self.server.next_available_port() if self.context.service_type(service_name) == "dropwizard" else port

                service_mapping_ports[service_mapping_name] = port

                orchestration_services[service_name] = {
                    "port": port,
                    "adminPort": admin_port,
                    "runFrom": run_from,
                    "classifier": classifier,
                    "version": version,
                    "appendArgs": append_args
                }

        return orchestration_services, service_mapping_ports


class SmStopRequest(SmRequest):

    def __init__(self, server, json_request_body, offlineMode):
        SmRequest.__init__(self, server, json_request_body, offlineMode)

    def process_request(self):

        if not self.server.is_running(self.test_id):
            raise BadRequestException("Invalid test id (or already stopped): %s" % self.test_id)

        self._log("Stopping test")

        drop_databases = self.json_body.get("dropDatabases", True)

        if type(drop_databases) is not bool:
            raise self._bad_request_exception("dropDatabases parameter must be boolean (value was: %s)" % drop_databases)

        errors = self._stop_services(drop_databases)

        if errors:
            self._log("Completed stopping services - errors occurred: %s" % str(errors))
            response.status = 500
            return json.dumps({"statusCode": 500, "errorMessage": errors})
        else:
            self._log("Successfully stopped services")
            response.status = 204


class SmShutdownRequest():

    def __init__(self, server):
        self.server = server

    def process_request(self):
        print "shutting down..."
        for test_id in self.server.running_tests:
            context = SmContext(self.server.application, test_id)
            context.log("Killing everything for testId %s..." % test_id)
            context.kill()
            for service_name in self.server.service_names_for_test(test_id):
                if context.service_data(service_name).get("hasMongo", False):
                    context.drop_database_for_service(service_name)

            context.drop_database_for_test()
            context.log("Successfully stopped all services for testId %s..." % test_id)

        print "finished shutting down..."


class SmServer:

    def __init__(self, application):

        self.application = application
        self.port_provider = PortProvider()

        # Map of test_id to list of service names
        self.running_tests = {}

    def next_available_port(self):
        return self.port_provider.next_available_port()

    def service_names_for_test(self, test_id):
        if self.is_running(test_id):
            return self.running_tests[test_id]
        else:
            return []

    def is_running(self, test_id):
        return test_id in self.running_tests

    def starting_test(self, test_id):
        if self.is_running(test_id):
            raise ServiceManagerException("Test '%s' is already running" % test_id)
        self.running_tests[test_id] = []

    def starting_service_for_test(self, test_id, service_name):
        if not self.is_running(test_id):
            raise ServiceManagerException("Test '%s' is not running" % test_id)
        self.running_tests[test_id] += [service_name]

    def test_stopped(self, test_id):
        del self.running_tests[test_id]


class SmVersionRequest:

    def __init__(self, server):
        self.application = server.application

    def process_request(self):
        service = request.query['service']
        if not service in self.application.service_mappings:
            raise BadRequestException("Service '%s' cannot be found in 'service_mappings.json', please update this file" % service)
        service_alias = self.application.service_mappings[service]
        if not service_alias in self.application.services:
            raise BadRequestException("Service '%s' cannot be found in 'services.json', please update this file" % service_alias)
        if not "versionEnv" in self.application.services[service_alias]:
            raise BadRequestException("'versionEnv' cannot be found for service '%s', please update 'services.json'" % service_alias)
        return {"variable": self.application.services[service_alias]["versionEnv"]}



#!/usr/bin/env python

from servicemanager.thirdparty.prettytable import PrettyTable
from servicemanager.service.smservice import SmServiceStatus
from servicemanager.smcontext import ServiceManagerException
from servicemanager.actions.colours import BColors

def _service_status_to_row(status):
    return [
        status.service_name,
        status.ppid,
        status.pid,
        status.uptime,
        status.mem,
        status.port,
        status.test_id,
        status.run_from,
        status.features,
        _format_healthcheck_status(status.healthcheck)
    ]


def _format_healthcheck_status(healthcheck):
    b = BColors()
    if healthcheck == SmServiceStatus.HEALTHCHECK_PASS:
        return b.bold + b.okgreen + "PASS" + b.endc
    elif healthcheck == SmServiceStatus.HEALTHCHECK_BOOT:
        return b.bold + b.warning + "BOOT" + b.endc
    elif healthcheck == SmServiceStatus.HEALTHCHECK_NONE:
        return b.okblue + "NONE" + b.endc
    else:
        raise ServiceManagerException("Unknown healthcheck status: %s" % healthcheck)


def dostatus(context, show_down_services):
    b = BColors()
    up_processes_table = PrettyTable()
    up_processes_table.field_names = ["name", "ppid", "pid", "uptime", "mem", "port", "test id", "run from", "features", "healthcheck"]
    up_processes_table.align = "r"
    up_processes_table.align["name"] = "l"
    up_processes_table.sortby = "name"
    up_processes_table.align["test id"] = "l"
    up_processes_table.align["run from"] = "l"
    up_processes_table.align["features"] = "l"

    down_processes_table = PrettyTable()
    down_processes_table.field_names = ["name", "healthcheck"]
    down_processes_table.align = "r"
    down_processes_table.align["name"] = "l"
    down_processes_table.sortby = "name"


    all_processes = context.process_manager.all_processes()
    for service in context.services():
        responses = service.status(all_processes)
        if responses:
            for response in responses:
                up_processes_table.add_row(_service_status_to_row(response))
        elif show_down_services:
            down_processes_table.add_row([service.service_name, b.bold + b.fail + "DOWN" + b.endc])

    result = ["Running:", str(up_processes_table)]
    if show_down_services:
        result.append("Down:")
        result.append(str(down_processes_table))
    return result


#!/usr/bin/env python
import os
import re

import requests
import types

from servicemanager.service.smservice import SmServiceStarter, SmService, SmServiceStatus
from servicemanager.smprocess import SmProcess, kill_pid
from servicemanager import subprocess


class SmExternalServiceStarter(SmServiceStarter):

    def process_arguments(self):
        pass

    def __init__(self, context, service_name, append_args):
        SmServiceStarter.__init__(self, context, service_name, "external", append_args)
        if not "cmd" in self.service_data:
            context.log("Could not start 'external' service '" + service_name + "', 'cmd' is missing from services.json")
        self.cmd = self.service_data["cmd"]

    def get_start_command(self, context=None):
        if self.append_args:
            if isinstance(self.append_args, types.ListType):
                return self.cmd + self.append_args
            else:
                return self.cmd
                self.log("WARNING: I was passed a non list for append args of '" + str(self.append_args) + "' I dont know what to do with this")
        else:
            return self.cmd

    def supports_append_args(self):
        return True

    def start(self):
        try:
            location = "."
            if "location" in self.service_data:
                location = self.service_data["location"]
            microservice_path = os.path.join(self.context.application.workspace, location)
            os.chdir(microservice_path)
            return subprocess.Popen(" ".join(self.get_start_command()), cwd=microservice_path, env=os.environ.copy(), shell=True).pid
        except Exception, e:
            self.log("Could not start service due to exception: " + str(e))


class SmExternalService(SmService):

    def __init__(self, context, service_name):
        SmService.__init__(self, context, service_name, "external")
        self.pattern = self.required_data("pattern")

    def get_pattern(self):
        return self.pattern

    def stop(self):
        processes = SmProcess.processes_matching(self.pattern)

        for process in processes:
            kill_pid(self.context, process.ppid)
            kill_pid(self.context, process.pid)
            print "name: %s\tppid: %s\tpid: %s\tuptime: %s" % (self.service_name, process.ppid, process.pid, process.uptime)

    def clean_up(self):
        pass

    def is_started_on_default_port(self):
        return len(SmProcess.processes_matching(self.pattern)) > 0

    def request_running_service_details_on_default_port(self):
        return None

    def run_healthcheck(self, port=None):
        if "healthcheck" in self.service_data:
            healthcheck_url = self.service_data["healthcheck"]["url"]
            healthcheck_response_regex = self.service_data["healthcheck"]["response"]

            try:
                ping_response = requests.get(healthcheck_url)
                response_text = ping_response.text
                return re.search(healthcheck_response_regex, response_text) is not None
            except requests.RequestException:
                return False
        else:
            return None

    def status(self, all_processes=None):
        processes = SmProcess.processes_matching(self.pattern, all_processes)

        if len(processes) == 0:
            return []

        if len(processes) > 1:
            self.log("POSSIBLE PROBLEM: Found more than one process")

        def _status_for_process(process):
            hc_result = self.run_healthcheck(process)
            healthcheck = SmServiceStatus.HEALTHCHECK_NONE
            if hc_result is not None:
                healthcheck = SmServiceStatus.HEALTHCHECK_PASS if self.run_healthcheck(process) else SmServiceStatus.HEALTHCHECK_BOOT
            return SmServiceStatus.for_process(self.service_name, process, "", "", "", "", healthcheck)


        return map(_status_for_process, processes)



#!/usr/bin/env python

import os

from servicemanager.subprocess import Popen
from servicemanager.smfile import force_chdir
from servicemanager.smnexus import SmNexus
from servicemanager.service.smservice import SmMicroServiceStarter
from smjvmservice import SmJvmService, SmJvmServiceStarter
from servicemanager.actions.colours import BColors

b = BColors()


class SmDropwizardServiceStarter(SmJvmServiceStarter):

    DEV_NULL = open(os.devnull, 'w')

    def __init__(self, context, service_name, run_from, port, admin_port, classifier, service_mapping_ports, version, proxy, append_args=None):
        SmMicroServiceStarter.__init__(self, context, service_name, "dropwizard", run_from, port, classifier, service_mapping_ports, version, proxy, append_args)

        self.admin_port = admin_port

        self.java_options = [
            "-Dfile.encoding=UTF8",
            "-Xmx64M",
            "-XX:+CMSClassUnloadingEnabled",
            "-XX:MaxPermSize=64m"
        ]

        self.java_home = context.application.java_home

    def _get_jar_filename(self):
        return self.context.get_jar_filename(self.service_name, self.run_from)

    def _get_binary_start_cmd(self):
        binary_data = self.service_data["binary"]
        filename = self._get_jar_filename()

        if "configurationFile" not in binary_data:
            print "ERROR: required config 'configurationFile' does not exist"
            return None

        configuration_file = binary_data["configurationFile"]
        output_configuration_file = os.path.join(self.run_from, configuration_file)
        microservice_target_path = self.context.get_microservice_target_path(self.service_name)
        _java_bin = os.path.join(self.java_home, os.path.join("bin", "java"))
        extra_params = self._build_extra_params()
        cmd = [_java_bin] + self.java_options + extra_params
        os.path.join(microservice_target_path, filename)
        return cmd + ["-jar",  microservice_target_path + filename, "server", microservice_target_path + output_configuration_file]


    def get_start_command(self, run_from):
        if run_from == "SOURCE":
            return self.sources["cmd"]
        else:
            return self._get_binary_start_cmd()

    def start_from_binary(self):

        microservice_target_path = self.context.get_microservice_target_path(self.service_name)
        force_chdir(microservice_target_path)

        if not self.context.offline:
            nexus = SmNexus(self.context, self.service_name)
            nexus.download_jar_if_necessary(self.run_from, self.version)

        filename = self._get_jar_filename()

        service_data = self.service_data
        binary_data = service_data["binary"]

        if "configurationFile" in binary_data:
            configuration_file = binary_data["configurationFile"]
        else:
            print "ERROR: required config 'configurationFile' does not exist"
            return None

        binary_to_run = os.path.join(microservice_target_path, filename)
        if os.path.exists(binary_to_run):
            force_chdir(self.run_from)
            os.system("jar xvf " + microservice_target_path + filename + " " + configuration_file + " > /dev/null")
            force_chdir(microservice_target_path)
            cmd = self.get_start_command("BINARY")
            process = Popen(cmd, env=os.environ.copy(), stdout=SmDropwizardServiceStarter.DEV_NULL, stderr=SmDropwizardServiceStarter.DEV_NULL, close_fds=True)
            if process.returncode == 1:
                print b.fail + "ERROR: could not start '" + self.service_name + "' " + b.endc
            return process.pid
        else:
            print "ERROR: the requested file: '" + binary_to_run + "' does not exist"
            return None

    def _build_extra_params(self):

        extra_params = [
            "-Ddw.http.port=%d" % self.port,
            "-Dservice.manager.serviceName=" + self.service_name
        ]
        extra_params += self.process_arguments()

        # Features are so specific - should this be in config?
        if self.context.features:
            for feature in self.context.features:
                extra_params += ["-Dfeature.%s=true" % feature]

        if self.admin_port:
            extra_params += ["-Ddw.http.adminPort=%d" % self.admin_port]

        if self.context.is_test:
            if self.service_data.get("hasMongo", False):
                extra_params += ["-Ddw.mongodb.db=%s-%s" % (self.context.database_name_prefix, self.service_name)]

        if self.service_mapping_ports and self.service_data.get("hasServiceMappings", False):
            for dependent_service_name in self.service_mapping_ports:
                extra_params += ["-Ddw.serviceMappings." + dependent_service_name + ".port=" + str(self.service_mapping_ports[dependent_service_name])]

        return extra_params

    def start_from_sources(self):

        sbt_extra_params = self._build_extra_params()

        base_dir = self.context.application.workspace + self.service_data["location"]
        cmd = self.sources["cmd"]

        process = None
        try:
            new_env = os.environ.copy()
            # SBT is so specific should this be in configuration?
            new_env["SBT_EXTRA_PARAMS"] = " ".join(sbt_extra_params)
            os.chdir(base_dir)
            process = Popen(self.get_start_command("SOURCE"), env=new_env, stdout=SmDropwizardServiceStarter.DEV_NULL, stderr=SmDropwizardServiceStarter.DEV_NULL, close_fds=True)
        except Exception, e:
            self.log("Could not start service in '%s' due to exception: %s" % (base_dir, str(e)))

        return process.pid


class SmDropwizardService(SmJvmService):

    def __init__(self, context, service_name):
        SmJvmService.__init__(self, context, service_name, "dropwizard")
        self.default_admin_port = self.required_data("defaultAdminPort")

    def clean_up(self):
        pass

    def post_stop(self):
        pass

    def get_details_url(self):
        return "http://localhost:${port}/ping/details"

    def get_port_argument(self):
        return "dw.http.port"

    def get_running_healthcheck_port(self, process):
        return process.extract_integer_argument('-Ddw.http.adminPort=(\d*)', self.default_admin_port)



#!/usr/bin/env python

from abc import abstractmethod

class SmServiceBase():
    def __init__(self, context, service_name, expected_service_type):
        self.context = context
        self.service_name = service_name
        self.service_data = self.context.service_data(service_name)
        self.service_type = context.service_type(service_name)

        if self.service_type != expected_service_type:
            raise context.exception("Cannot construct '%s' as type '%s', services.json states its type is: '%s'" % (service_name, expected_service_type, self.service_type))

    def log(self, message):
        self.context.log("[%s]: %s" % (self.service_name, message))

    def required_data(self, key):
        value = self.service_data.get(key, None)

        if not value:
            message = "Missing required '%s' configuration in services.json" % key
            self.log(message)
            raise self.context.exception(message)

        return value


class SmService(SmServiceBase):

    def __init__(self, context, service_name, expected_service_type):
        SmServiceBase.__init__(self, context, service_name, expected_service_type)

    @abstractmethod
    def is_started_on_default_port(self):
        pass

    @abstractmethod
    def request_running_service_details_on_default_port(self):
        pass

    @abstractmethod
    def status(self, all_processes):
        pass

    @abstractmethod
    def run_healthcheck(self, process):
        pass

    @abstractmethod
    def get_pattern(self):
        pass


class SmServiceStarter(SmServiceBase):

    def __init__(self, context, service_name, expected_service_type, append_args):
        SmServiceBase.__init__(self, context, service_name, expected_service_type)
        self.append_args = append_args

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def process_arguments(self):
        pass

    @abstractmethod
    def supports_append_args(self):
        return False

    @abstractmethod
    def get_start_command(self, context = None):
        return ["get_start_command() not implemented for this type of service - fork and make a pull request :)"]


class SmMicroServiceStarter(SmServiceStarter):

    def __init__(self, context, service_name, expected_service_type, run_from, port, classifier, service_mapping_ports, version, proxy, append_args):
        SmServiceStarter.__init__(self, context, service_name, expected_service_type, append_args)

        self.run_from = run_from
        self.version = version
        self.classifier = classifier
        self.service_mapping_ports = service_mapping_ports
        self.port = port if port else self.service_data["defaultPort"]
        self.proxy = proxy

        if "sources" not in self.service_data:
            raise context.exception("Invalid services.json entry for %s service '%s', missing 'sources'" % (self.service_type, self.service_name))

        self.sources = self.service_data["sources"]

    @abstractmethod
    def start(self, appendArgs=None):
        pass

    @abstractmethod
    def process_arguments(self):
        pass


class SmServiceStatus():

    HEALTHCHECK_NONE = "NONE"
    HEALTHCHECK_PASS = "PASS"
    HEALTHCHECK_BOOT = "BOOT"

    def __init__(self, service_name, ppid, pid, uptime, mem, port, test_id, run_from, features, healthcheck):
        self.service_name = service_name
        self.ppid = ppid
        self.pid = pid
        self.uptime = uptime
        self.mem = mem
        self.port = port
        self.test_id = test_id
        self.run_from = run_from
        self.features = features
        self.healthcheck = healthcheck

    @staticmethod
    def for_process(service_name, process, port, test_id, run_from, features, healthcheck):
        return SmServiceStatus(service_name, process.ppid, process.pid, process.uptime, process.mem, port, test_id, run_from, features, healthcheck)


import os
import sys
from smnexus import SmNexus
from smbintray import SmBintray


class SmArtifactRepoFactory():

    @staticmethod
    def get_repository(context, service_name, service_binary_config):
    	if SmArtifactRepoFactory._is_nexus(service_binary_config):
            return SmNexus(context, service_name)
        else:
            return SmBintray(context, service_name)

    @staticmethod
    def get_play_app_extension(service_binary_config):
        if SmArtifactRepoFactory._is_nexus(service_binary_config):
            return ".zip"
        else:
            return ".tgz"

    @staticmethod
    def _is_nexus(service_binary_config):
        if "nexus" in service_binary_config:
            return True
        else:
        	return False    

